#include "Multiplexer.h"
#include <vector>

Multiplexer::Multiplexer(byte LATCHpin, byte CLRpin, LiquidCrystal& lcd):
  _chainLength(17),
  _LATCHpin(LATCHpin),
  _CLRpin(CLRpin),
  _HV20220SPISettings(SPISettings(1000000, MSBFIRST, SPI_MODE0)),
  _lcd(lcd) {
    for (unsigned int i=0; i<_chainLength; i++) {
      _outputValues.push_back(0x00); // Initialize to all switches off.
      _savedValues.push_back(0x00);
    }
  }

void Multiplexer::begin(void) {
  pinMode(_LATCHpin, OUTPUT);
  pinMode(_CLRpin, OUTPUT);

  digitalWrite(_LATCHpin, HIGH);
  digitalWrite(_CLRpin, LOW);
  
  setLoad(1, MUX_1k);
  setLoad(2, MUX_1k);
  setLoad(3, MUX_1k);
  setLoad(4, MUX_1k);
  
  EnableOutput(); // Enable output through any load.
}

void Multiplexer::SendCommand(std::vector<byte> &data) {
  // Configure the SPI Port for this chip.
  SPI.beginTransaction(_HV20220SPISettings);
  
  // Write the data out.
  for (std::vector<byte>::iterator i=data.begin(); i != data.end(); ++i) {
    // For this we will ignore the returned data.
    SPI.transfer(*i);
  }
  
  // Latch in the new data to the actual switches.
  delayMicroseconds(1);
  digitalWrite(_LATCHpin, LOW);
  delayMicroseconds(1);
  digitalWrite(_LATCHpin, HIGH);
  
  SPI.endTransaction();
}

void Multiplexer::setChip(byte channel, byte configuration) {
  // Check if outputs are enabled.
  if (_outputenabled) {
    _outputValues[channel] = configuration;
    _savedValues[channel] = configuration;
    sendUpdate();
  }
  else {
    _savedValues[channel] = configuration;
    if ((channel == 16) or (channel == 12) or (channel == 8) or (channel == 4) or (channel == 0)) {
      // If the channel is a load or the AC/DC switch, allow it to set the configuration.
      // Do not change outputValue if it is an electrode.
      _outputValues[channel] = configuration;
      sendUpdate();
    }
  }
}

void Multiplexer::sendUpdate(void) {
  SendCommand(_outputValues);
}

void Multiplexer::Enable(void) {
  digitalWrite(_CLRpin, LOW);
}

void Multiplexer::Disable(void) {
  digitalWrite(_CLRpin, HIGH);
}

void Multiplexer::EnableOutput(void) {
  _outputenabled = true;
  _outputValues = _savedValues; // Copy back the saved multiplexer setting.
  updateElectrodeLCD();
  sendUpdate();
}

void Multiplexer::DisableOutput(void) {  
  setLoad(0, MUX_10k); // Set all loads to the 10k.
  setLoad(1, MUX_10k);
  setLoad(2, MUX_10k);
  setLoad(3, MUX_10k);

  setElectrode(0, 0); // For each channel, disconnect all electrodes.
  setElectrode(1, 0);
  setElectrode(2, 0);
  setElectrode(3, 0);
  
  _outputenabled = false;
  
  sendUpdate();
}

void Multiplexer::setElectrode(byte channel, byte electrode) {
  byte target_chip, code;
  
  if (electrode > 0 and electrode <= 24) {
    // This calculates the target chip for any electrode and channel combination.
    target_chip = (_chainLength-2)-((channel-1)*4+floor(((float)electrode-1.0)/8.0));

    // This calculates the electrode code for the target chip.
    code = 1<<((electrode-1)%8); 
  }
  else if (electrode == 0) {
    target_chip = 0; 
    code = 0;
  }
  else {
    SerialUSB.println("ERROR: Electrode/Load Setting Invalid.");
    return;
  }

  // In the data array, clear the electrode switches for the given channel first.
  switch(channel) {
    case(1) :
      _outputValues[15] = 0;
      _outputValues[14] = 0;
      _outputValues[13] = 0;
      if (electrode == 0) target_chip = 15;
      break;
    case(2) :
      _outputValues[11] = 0;
      _outputValues[10] = 0;
      _outputValues[9] = 0;
      if (electrode == 0) target_chip = 11;
      break;
    case(3) :
      _outputValues[7] = 0;
      _outputValues[6] = 0;
      _outputValues[5] = 0;
      if (electrode == 0) target_chip = 7;
      break;
    case(4) :
      _outputValues[3] = 0;
      _outputValues[2] = 0;
      _outputValues[1] = 0;
      if (electrode == 0) target_chip = 3;
      break;
    default :
      SerialUSB.println("ERROR: Attempted to set invalid channel electrode configuration.");
      return;
  }
  
  if (electrode >= 0 and electrode <= 24) {
    setChip(target_chip, code);
    _savedElectrode[channel-1] = electrode;
    updateElectrodeLCD();
  }
}

// This function is a legacy emulator to match the old board load numbers (0 = no load, 1 = 10k, 2 = 10nF, 3 = GND).

void Multiplexer::setLegacyLoad(byte channel, byte load) {
  byte target_chip;
  switch(channel) {
    case(1) :
      target_chip = 12;
      break;
    case(2) :
      target_chip = 8;
      break;
    case(3) :
      target_chip = 4;
      break;
    case(4) :
      target_chip = 0;
      break;
    default:
      SerialUSB.println("ERROR: Attempted to set invalid load channel.");
      return;
  }

  byte code;
  switch(load) {
    case(0) : // This means connect to nothing.
      code = MUX_OFF;
      break;
    case(1) : // This means connect to 10k resistor.
      code = MUX_10k;
      break;
    case(2) : // This means connect to 10nF capacitor.
      code = MUX_10nF;
      break;
    case(3) : // This means to ground the connection.
      code = MUX_GND;
  }

  // And actually set the corresponding chip configuration.
  setChip(target_chip, code);
  _load[channel-1] = load;
  updateLoadLCD();
}

void Multiplexer::setLoad(byte channel, byte load) {
  byte target_chip;
  switch(channel) {
    case(1) :
      target_chip = 12;
      break;
    case(2) :
      target_chip = 8;
      break;
    case(3) :
      target_chip = 4;
      break;
    case(4) :
      target_chip = 0;
      break;
    default:
      SerialUSB.println("ERROR: Attempted to set invalid load channel.");
      return;
  }

  // In this general case, assume that the caller is using the true output load code for the desired configuration.
  // This will only work on X3 boards, and should not be called by any old X2 commands.

  setChip(target_chip, load);
  _load[channel-1] = load;
  updateLoadLCD();
}

void Multiplexer::updateElectrodeLCD() {
  // Check if outputs are enabled.
  if (_outputenabled) {
    for (byte i=0; i<4; i++) {
      _electrode[i] = _savedElectrode[i];
    }
    char buf[40];
    sprintf(buf, "CH1: %02d   CH2: %02d   CH3: %02d   CH4: %02d", _electrode[0], _electrode[1], _electrode[2], _electrode[3]);
    _lcd.setCursor(0, 0);
    _lcd.print(buf);
  }
}

void Multiplexer::updateLoadLCD() {
  char buf[40];
  sprintf(buf, "LD1: %02X   LD2: %02X   LD3: %02X   LD4: %02X", _load[0], _load[1], _load[2], _load[3]);
  _lcd.setCursor(0, 1);
  _lcd.print(buf);
}

void Multiplexer::DCMode(void) {
  setChip(16, 0xFF);
}

void Multiplexer::ACMode(void) {
  setChip(16, 0x00);
}
