#pragma once

#include <Arduino.h>
#include <SPI.h>
#include <LiquidCrystal.h>
#include <vector>

// Note that the multiplexer library is *specific* to this configuration. The devices are fixed
// and daisy-chained, so are in reverse order with the first chip in the chain having the last
// address. Bit order per chip is assumed to be MSB first, so setting 0b00000001 will correspond
// to activating the connection SW0 on the given chip.

// Hardware modes defined by HV20220PJ datasheet.

/* For this multiplexer, the daisy chain ordering is:
 * 16 - ACDC switching mux for current stimulator.
 * 15 - CH1 Electrodes 01-08
 * 14 - CH1 Electrodes 09-16
 * 13 - CH1 Electrodes 17-24
 * 12 - CH1 Loads (0 ohm, 10 ohm, 100 ohm, 1kohm, 10kohm, 100kohm, 1Mohm, 10nF)
 * 11 - CH2 Electrodes 01-08
 * 10 - CH2 Electrodes 09-16
 * 9 - CH2 Electrodes 17-24
 * 8 - CH2 Loads (0 ohm, 10 ohm, 100 ohm, 1kohm, 10kohm, 100kohm, 1Mohm, 10nF)
 * 7 - CH3 Electrodes 01-08
 * 6 - CH3 Electrodes 09-16
 * 5 - CH3 Electrodes 17-24
 * 4 - CH3 Loads (0 ohm, 10 ohm, 100 ohm, 1kohm, 10kohm, 100kohm, 1Mohm, 10nF)
 * 3 - CH4 Electrodes 01-08
 * 2 - CH4 Electrodes 09-16
 * 1 - CH4 Electrodes 17-24
 * 0 - CH4 Loads (0 ohm, 10 ohm, 100 ohm, 1kohm, 10kohm, 100kohm, 1Mohm, 10nF)
 */

// Load definitions for easier use.

#define MUX_OFF 0b00000000
#define MUX_GND 0b00000001
#define MUX_10ohm 0b00000010
#define MUX_100ohm 0b00000100
#define MUX_1k 0b00001000
#define MUX_10k 0b00010000
#define MUX_100k 0b00100000
#define MUX_1M 0b01000000
#define MUX_10nF 0b10000000
#define MUX_10ohm_10nF 0b10000010
#define MUX_100ohm_10nF 0b10000100
#define MUX_1k_10nF 0b10001000
#define MUX_10k_10nF 0b10010000
#define MUX_100k_10nF 0b10100000
#define MUX_1M_10nF 0b11000000

class Multiplexer {
  private:
    byte _chainLength, _LATCHpin, _CLRpin;
    SPISettings _HV20220SPISettings;
    LiquidCrystal& _lcd;
    std::vector<byte> _outputValues;
    std::vector<byte> _savedValues;
    byte _electrode[4], _savedElectrode[4], _load[4];
    boolean _outputenabled;
    void SendCommand(std::vector<byte> &data);
  public:
    Multiplexer(byte LATCHpin, byte CLRpin, LiquidCrystal& lcd);
    void begin(void);
    void setChip(byte channel, byte configuration);
    void setElectrode(byte channel, byte electrode);
    void setLegacyLoad(byte channel, byte load);
    void setLoad(byte channel, byte load);
    void Enable(void);
    void Disable(void);
    void DisableOutput(void);
    void EnableOutput(void);
    void updateElectrodeLCD();
    void updateLoadLCD();
    void DCMode(void);
    void ACMode(void);
    void sendUpdate(void);
};
